@@ -48,7 +48,6 @@
 #endif
 
 #include <stdio.h>
-#define NEED_EVENTS
 #include <X11/X.h>
 #include <X11/Xproto.h>
 
@@ -68,9 +67,19 @@
 #include "xf86OSmouse.h"
 #include "compiler.h"
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+#include <xserver-properties.h>
+#include "exevents.h"
+#endif
+
 #include "xisb.h"
 #include "mipointer.h"
 
+#ifndef HAVE_XORG_SERVER_1_5_0
+#include <xf86_ansic.h>
+#include <xf86_libc.h>
+#endif
+
 /*****************************************************************************
  *	Local Headers
  ****************************************************************************/
@@ -93,56 +102,63 @@
 /*
  * Version constants
  */
-#define VMMOUSE_MAJOR_VERSION 12
-#define VMMOUSE_MINOR_VERSION 5
-#define VMMOUSE_PATCHLEVEL 1
 #define VMMOUSE_DRIVER_VERSION \
-   (VMMOUSE_MAJOR_VERSION * 65536 + VMMOUSE_MINOR_VERSION * 256 + VMMOUSE_PATCHLEVEL)
+   (PACKAGE_VERSION_MAJOR * 65536 + PACKAGE_VERSION_MINOR * 256 + PACKAGE_VERSION_PATCHLEVEL)
 #define VMMOUSE_DRIVER_VERSION_STRING \
-    VMW_STRING(VMMOUSE_MAJOR_VERSION) "." VMW_STRING(VMMOUSE_MINOR_VERSION) \
-    "." VMW_STRING(VMMOUSE_PATCHLEVEL)
+    VMW_STRING(PACKAGE_VERSION_MAJOR) "." VMW_STRING(PACKAGE_VERSION_MINOR) \
+    "." VMW_STRING(PACKAGE_VERSION_PATCHLEVEL)
 
 /*
  * Standard four digit version string expected by VMware Tools installer.
- * As the driver's version is only  {major, minor, patchlevel}, simply append an
- * extra zero for the fourth digit.
+ * As the driver's version is only  {major, minor, patchlevel},
+ * The fourth digit may describe the commit number relative to the
+ * last version tag as output from `git describe`
  */
 #ifdef __GNUC__
+#ifdef VMW_SUBPATCH
+const char vm_mouse_version[] __attribute__((section(".modinfo"),unused)) =
+    "version=" VMMOUSE_DRIVER_VERSION_STRING "." VMW_STRING(VMW_SUBPATCH);
+#else
 const char vm_mouse_version[] __attribute__((section(".modinfo"),unused)) =
     "version=" VMMOUSE_DRIVER_VERSION_STRING ".0";
+#endif /*VMW_SUBPATCH*/
 #endif
 
 
 /*****************************************************************************
  *	static function header
  ****************************************************************************/
-#ifdef XFree86LOADER
-static const OptionInfoRec *VMMouseAvailableOptions(void *unused);
-#endif
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+static int VMMousePreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
+#else
 static InputInfoPtr VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags);
+static void VMMouseCloseProc(InputInfoPtr pInfo);
+static Bool VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
+			       int v3, int v4, int v5, int *x, int *y);
+#endif
 static void VMMouseUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
 static void MouseCommonOptions(InputInfoPtr pInfo);
 static void GetVMMouseMotionEvent(InputInfoPtr pInfo);
 static void VMMousePostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy, int dz, int dw);
 static void VMMouseDoPostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy);
 static Bool VMMouseDeviceControl(DeviceIntPtr device, int mode);
-static void VMMouseCloseProc(LocalDevicePtr local);
-static int  VMMouseControlProc(LocalDevicePtr local, xDeviceCtl * control);
+static int  VMMouseControlProc(InputInfoPtr pInfo, xDeviceCtl * control);
 static void VMMouseReadInput(InputInfoPtr pInfo);
 static int  VMMouseSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode);
-static Bool VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
-			       int v3, int v4, int v5, int *x, int *y);
 static void MouseCtrl(DeviceIntPtr device, PtrCtrl *ctrl);
 
 /******************************************************************************
  *		Definitions
  *****************************************************************************/
 typedef struct {
-   int 		screenNum;
-   Bool 	vmmouseAvailable;
-   Bool		relative;
+   int                 screenNum;
+   Bool                vmmouseAvailable;
+   VMMOUSE_INPUT_DATA  vmmousePrevInput;
+   Bool                isCurrRelative;
+   Bool                absoluteRequested;
 } VMMousePrivRec, *VMMousePrivPtr;
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 7
 static const char *reqSymbols[] = {
    "InitPointerDeviceStruct",
    "LoaderSymbol",
@@ -189,6 +205,7 @@ static const char *reqSymbols[] = {
    "Xstrdup",
    NULL
 };
+#endif
 
 InputDriverRec VMMOUSE = {
    1,
@@ -196,70 +213,11 @@ InputDriverRec VMMOUSE = {
    NULL,
    VMMousePreInit,
    VMMouseUnInit,
-   NULL,
+   NULL
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 11
+       ,
    0
-};
-
-typedef enum {
-    OPTION_ALWAYS_CORE,
-    OPTION_SEND_CORE_EVENTS,
-    OPTION_CORE_POINTER,
-    OPTION_SEND_DRAG_EVENTS,
-    OPTION_HISTORY_SIZE,
-    OPTION_DEVICE,
-    OPTION_PROTOCOL,
-    OPTION_BUTTONS,
-    OPTION_EMULATE_3_BUTTONS,
-    OPTION_EMULATE_3_TIMEOUT,
-    OPTION_CHORD_MIDDLE,
-    OPTION_FLIP_XY,
-    OPTION_INV_X,
-    OPTION_INV_Y,
-    OPTION_ANGLE_OFFSET,
-    OPTION_Z_AXIS_MAPPING,
-    OPTION_SAMPLE_RATE,
-    OPTION_RESOLUTION,
-    OPTION_EMULATE_WHEEL,
-    OPTION_EMU_WHEEL_BUTTON,
-    OPTION_EMU_WHEEL_INERTIA,
-    OPTION_X_AXIS_MAPPING,
-    OPTION_Y_AXIS_MAPPING,
-    OPTION_AUTO_SOFT,
-    OPTION_DRAGLOCKBUTTONS
-} MouseOpts;
-
-/*
- * Define the acceptable mouse options
- * Currently not all of those options are supported
- *
- */
-static const OptionInfoRec mouseOptions[] = {
-    { OPTION_ALWAYS_CORE,	"AlwaysCore",	  OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_SEND_CORE_EVENTS,	"SendCoreEvents", OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_CORE_POINTER,	"CorePointer",	  OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_SEND_DRAG_EVENTS,	"SendDragEvents", OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_HISTORY_SIZE,	"HistorySize",	  OPTV_INTEGER,	{0}, FALSE },
-    { OPTION_DEVICE,		"Device",	  OPTV_STRING,	{0}, FALSE },
-    { OPTION_PROTOCOL,		"Protocol",	  OPTV_STRING,	{0}, FALSE },
-    { OPTION_BUTTONS,		"Buttons",	  OPTV_INTEGER,	{0}, FALSE },
-    { OPTION_EMULATE_3_BUTTONS,	"Emulate3Buttons",OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_EMULATE_3_TIMEOUT,	"Emulate3Timeout",OPTV_INTEGER,	{0}, FALSE },
-    { OPTION_CHORD_MIDDLE,	"ChordMiddle",	  OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_FLIP_XY,		"FlipXY",	  OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_INV_X,		"InvX",		  OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_INV_Y,		"InvY",		  OPTV_BOOLEAN,	{0}, FALSE },
-    { OPTION_ANGLE_OFFSET,	"AngleOffset",	  OPTV_INTEGER,	{0}, FALSE },
-    { OPTION_Z_AXIS_MAPPING,	"ZAxisMapping",	  OPTV_STRING,	{0}, FALSE },
-    { OPTION_SAMPLE_RATE,	"SampleRate",	  OPTV_INTEGER,	{0}, FALSE },
-    { OPTION_RESOLUTION,	"Resolution",	  OPTV_INTEGER,	{0}, FALSE },
-    { OPTION_EMULATE_WHEEL,	"EmulateWheel",	  OPTV_BOOLEAN, {0}, FALSE },
-    { OPTION_EMU_WHEEL_BUTTON,	"EmulateWheelButton", OPTV_INTEGER, {0}, FALSE },
-    { OPTION_EMU_WHEEL_INERTIA,	"EmulateWheelInertia", OPTV_INTEGER, {0}, FALSE },
-    { OPTION_X_AXIS_MAPPING,	"XAxisMapping",	  OPTV_STRING,	{0}, FALSE },
-    { OPTION_Y_AXIS_MAPPING,	"YAxisMapping",	  OPTV_STRING,	{0}, FALSE },
-    { OPTION_AUTO_SOFT,		"AutoSoft",	  OPTV_BOOLEAN, {0}, FALSE },
-    { OPTION_DRAGLOCKBUTTONS,	"DragLockButtons",OPTV_STRING,	{0}, FALSE },
-    { -1,			NULL,		  OPTV_NONE,	{0}, FALSE }
+#endif
 };
 
 static char reverseMap[32] = { 0,  4,  2,  6,  1,  5,  3,  7,
@@ -269,6 +227,56 @@ static char reverseMap[32] = { 0,  4,  2,  6,  1,  5,  3,  7,
 
 #define reverseBits(map, b)	(((b) & ~0x0f) | map[(b) & 0x0f])
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+static int
+VMMouseInitPassthru(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+{
+   InputAttributes *attrs = NULL;
+   InputOption *input_options = NULL, *tmp, *opts;
+   pointer options;
+   DeviceIntPtr dev;
+   int rc;
+
+   options = xf86OptionListDuplicate(pInfo->options);
+   options = xf86ReplaceStrOption(options, "Driver", "mouse");
+
+   while(options) {
+      tmp = calloc(1, sizeof(InputOption));
+      tmp->key = xf86OptionName(options);
+      tmp->value = xf86OptionValue(options);
+      tmp->next = input_options;
+      input_options = tmp;
+      options = xf86NextOption(options);
+   }
+
+   rc = NewInputDeviceRequest(input_options, attrs, &dev);
+
+   opts = input_options;
+   tmp = opts;
+   while(opts) {
+      tmp = opts->next;
+      free(opts->key);
+      free(opts->value);
+      free(opts);
+      opts = tmp;
+   }
+
+   return rc;
+}
+
+#else /* if ABI_XINPUT_VERSION < 12 */
+static InputInfoPtr
+VMMouseInitPassthru(InputDriverPtr drv, IDevPtr dev, int flags)
+{
+   InputDriverRec *passthruMouse;
+   passthruMouse = (InputDriverRec *)LoaderSymbol("MOUSE");
+   if(passthruMouse != NULL) {
+      return (passthruMouse->PreInit)(drv, dev, flags);
+   } else {
+      return NULL;
+   }
+}
+#endif
 
 /*
  *----------------------------------------------------------------------
@@ -289,12 +297,17 @@ static char reverseMap[32] = { 0,  4,  2,  6,  1,  5,  3,  7,
  *----------------------------------------------------------------------
  */
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
+static int
+VMMouseNewPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
+
 static InputInfoPtr
 VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 {
    InputInfoPtr pInfo;
-   MouseDevPtr pMse;
-   VMMousePrivPtr mPriv;
+
+#ifndef NO_MOUSE_MODULE
+{
    OSMouseInfoPtr osInfo = NULL;
 
    /*
@@ -303,13 +316,9 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    osInfo = xf86OSMouseInit(0);
    if (!osInfo)
       return FALSE;
+}
+#endif
 
-   mPriv = xcalloc (1, sizeof (VMMousePrivRec));
-
-
-   if (!mPriv) {
-      return NULL;
-   }
    /*
     * try to enable vmmouse here
     */
@@ -318,22 +327,12 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
        * vmmouse failed
        * Fall back to normal mouse module
        */
-      InputDriverRec *passthruMouse;
       xf86Msg(X_ERROR, "VMWARE(0): vmmouse enable failed\n");
-      mPriv->vmmouseAvailable = FALSE;
-      passthruMouse = (InputDriverRec *)LoaderSymbol("MOUSE");
-      xfree(mPriv);
-      if(passthruMouse != NULL){
-	 return (passthruMouse->PreInit)(drv, dev, flags);
-      } else {
-	 return NULL;
-      }
-
+      return VMMouseInitPassthru(drv, dev, flags);
    } else {
       /*
        * vmmouse is available
        */
-      mPriv->vmmouseAvailable = TRUE;
       xf86Msg(X_INFO, "VMWARE(0): vmmouse is available\n");
       /*
        * Disable the absolute pointing device for now
@@ -343,22 +342,14 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    }
 
    if (!(pInfo = xf86AllocateInput(drv, 0))) {
-      xfree(mPriv);
       return NULL;
    }
 
-   /* Settup the pInfo */
    pInfo->name = dev->identifier;
-   pInfo->type_name = XI_MOUSE;
-   pInfo->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
-   pInfo->device_control = VMMouseDeviceControl;
-   pInfo->read_input = VMMouseReadInput;
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
    pInfo->motion_history_proc = xf86GetMotionEvents;
 #endif
-   pInfo->control_proc = VMMouseControlProc;
    pInfo->close_proc = VMMouseCloseProc;
-   pInfo->switch_mode = VMMouseSwitchMode;
    pInfo->conversion_proc = VMMouseConvertProc;
    pInfo->reverse_conversion_proc = NULL;
    pInfo->fd = -1;
@@ -366,11 +357,62 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    pInfo->private_flags = 0;
    pInfo->always_core_feedback = 0;
    pInfo->conf_idev = dev;
+   pInfo->flags = XI86_POINTER_CAPABLE | XI86_SEND_DRAG_EVENTS;
+
+   /* Collect the options, and process the common options. */
+   xf86CollectInputOptions(pInfo, NULL, NULL);
+   xf86ProcessCommonOptions(pInfo, pInfo->options);
+
+   if (VMMouseNewPreInit(drv, pInfo, flags) == Success)
+       pInfo->flags |= XI86_CONFIGURED;
+
+   return pInfo;
+}
+
+static int
+VMMouseNewPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+#else /* if ABI_XINPUT_VERSION >= 12 */
+static int
+VMMousePreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
+#endif
+{
+   MouseDevPtr pMse = NULL;
+   VMMousePrivPtr mPriv = NULL;
+   int rc = Success;
+
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+   /* For ABI < 12, we need to return the wrapped driver's pInfo (see
+    * above). ABI 12, we call NIDR and are done */
+   if (!VMMouseClient_Enable()) {
+      xf86Msg(X_ERROR, "VMWARE(0): vmmouse enable failed\n");
+      return VMMouseInitPassthru(drv, pInfo, flags);
+   } else {
+      xf86Msg(X_INFO, "VMWARE(0): vmmouse is available\n");
+      VMMouseClient_Disable();
+   }
+#endif
+
+   mPriv = calloc (1, sizeof (VMMousePrivRec));
+
+   if (!mPriv) {
+      rc = BadAlloc;
+      goto error;
+   }
+
+   mPriv->absoluteRequested = FALSE;
+   mPriv->vmmouseAvailable = TRUE;
+
+   /* Settup the pInfo */
+   pInfo->type_name = XI_MOUSE;
+   pInfo->device_control = VMMouseDeviceControl;
+   pInfo->read_input = VMMouseReadInput;
+   pInfo->control_proc = VMMouseControlProc;
+   pInfo->switch_mode = VMMouseSwitchMode;
 
    /* Allocate the MouseDevRec and initialise it. */
-   if (!(pMse = xcalloc(sizeof(MouseDevRec), 1))) {
-      xfree(mPriv);
-      return pInfo;
+   if (!(pMse = calloc(sizeof(MouseDevRec), 1))) {
+      rc = BadAlloc;
+      goto error;
    }
 
    pInfo->private = pMse;
@@ -380,10 +422,6 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    pMse->mousePriv = mPriv;
 
 
-   /* Collect the options, and process the common options. */
-   xf86CollectInputOptions(pInfo, NULL, NULL);
-   xf86ProcessCommonOptions(pInfo, pInfo->options);
-
    /* Check if the device can be opened. */
    pInfo->fd = xf86OpenSerial(pInfo->options);
    if (pInfo->fd == -1) {
@@ -391,11 +429,8 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 	 xf86Msg(X_WARNING, "%s: cannot open input device\n", pInfo->name);
       else {
 	 xf86Msg(X_ERROR, "%s: cannot open input device\n", pInfo->name);
-	 if (pMse->mousePriv)
-	    xfree(pMse->mousePriv);
-	 xfree(pMse);
-	 pInfo->private = NULL;
-	 return pInfo;
+	 rc = BadValue;
+	 goto error;
       }
    }
    xf86CloseSerial(pInfo->fd);
@@ -407,17 +442,17 @@ VMMousePreInit(InputDriverPtr drv, IDevPtr dev, int flags)
    /* set up the current screen num */
    mPriv->screenNum = xf86SetIntOption(pInfo->options, "ScreenNumber", 0);
 
-   pInfo->flags |= XI86_CONFIGURED;
-   return pInfo;
-}
+   return Success;
 
-#ifdef XFree86LOADER
-static const OptionInfoRec *
-VMMouseAvailableOptions(void *unused)
-{
-    return (mouseOptions);
+error:
+   pInfo->private = NULL;
+   if (mPriv)
+      free(mPriv);
+   if (pMse)
+      free(pMse);
+
+   return rc;
 }
-#endif
 
 
 /*
@@ -476,6 +511,7 @@ VMMouseDoPostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy)
     VMMousePrivPtr mPriv;
     int truebuttons;
     int id, change;
+    Bool mouseMoved = FALSE;
 
     pMse = pInfo->private;
     mPriv = (VMMousePrivPtr)pMse->mousePriv;
@@ -489,7 +525,14 @@ VMMouseDoPostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy)
 
     buttons = reverseBits(reverseMap, buttons);
 
-    if (dx || dy) {
+    if (mPriv->isCurrRelative) {
+       mouseMoved = dx || dy;
+    } else {
+       mouseMoved = (dx != mPriv->vmmousePrevInput.X) ||
+                    (dy != mPriv->vmmousePrevInput.Y) ||
+                    (mPriv->vmmousePrevInput.Flags & VMMOUSE_MOVE_RELATIVE);
+    }
+    if (mouseMoved) {
 
 #ifdef CALL_CONVERSION_PROC
         /*
@@ -498,7 +541,7 @@ VMMouseDoPostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy)
          */
         VMMouseConvertProc(pInfo, 0, 2, dx, dy, 0, 0, 0, 0, &dx, &dy);
 #endif
-        xf86PostMotionEvent(pInfo->dev, !mPriv->relative, 0, 2, dx, dy);
+        xf86PostMotionEvent(pInfo->dev, !mPriv->isCurrRelative, 0, 2, dx, dy);
     }
 
     if (truebuttons != pMse->lastButtons) {
@@ -546,7 +589,7 @@ VMMousePostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy, int dz, int dw
 	break;
     case MSE_MAPTOX:
 	if (dz != 0) {
-	   if(mPriv->relative)
+	   if(mPriv->isCurrRelative)
 	      dx = dz;
 	   else
 	      dx += dz;
@@ -555,7 +598,7 @@ VMMousePostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy, int dz, int dw
 	break;
     case MSE_MAPTOY:
 	if (dz != 0) {
-	   if(mPriv->relative)
+	   if(mPriv->isCurrRelative)
 	      dy = dz;
 	   else
 	      dy += dz;
@@ -590,7 +633,7 @@ VMMousePostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy, int dz, int dw
      */
     if (zbutton) {
 	buttons &= ~zbutton;
-	if(mPriv->relative)
+	if(mPriv->isCurrRelative)
 	   VMMouseDoPostEvent(pInfo, buttons, 0, 0);
 	else
 	   VMMouseDoPostEvent(pInfo, buttons, dx, dy);
@@ -603,8 +646,7 @@ VMMousePostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy, int dz, int dw
  *
  * FlushButtons --
  *
- * 	FlushButtons -- send button up events for sanity. It is called
- *	during DEVICE_ON in VMMouseDeviceControl
+ * 	FlushButtons -- reset button states.
  *
  * Results:
  * 	None
@@ -618,18 +660,7 @@ VMMousePostEvent(InputInfoPtr pInfo, int buttons, int dx, int dy, int dz, int dw
 static void
 FlushButtons(MouseDevPtr pMse)
 {
-
-    /* If no button down is pending xf86PostButtonEvent()
-     * will discard them. So we are on the safe side. */
-
-    int i, blocked;
-
     pMse->lastButtons = 0;
-
-    blocked = xf86BlockSIGIO ();
-    for (i = 1; i <= 5; i++)
-	xf86PostButtonEvent(pMse->device,0,i,0,0,0);
-    xf86UnblockSIGIO (blocked);
 }
 
 
@@ -675,7 +706,7 @@ MouseCommonOptions(InputInfoPtr pInfo)
    /*
     * Process option for ZAxisMapping
     */
-   s = xf86SetStrOption(pInfo->options, "ZAxisMapping", NULL);
+   s = xf86SetStrOption(pInfo->options, "ZAxisMapping", "4 5");
    if (s) {
       int b1 = 0, b2 = 0, b3 = 0, b4 = 0;
       char *msg = NULL;
@@ -709,7 +740,7 @@ MouseCommonOptions(InputInfoPtr pInfo)
       }
       if (msg) {
 	 xf86Msg(X_CONFIG, "%s: ZAxisMapping: %s\n", pInfo->name, msg);
-	 xfree(msg);
+	 free(msg);
       } else {
 	 xf86Msg(X_WARNING, "%s: Invalid ZAxisMapping value: \"%s\"\n",
 		 pInfo->name, s);
@@ -742,7 +773,7 @@ MouseCommonOptions(InputInfoPtr pInfo)
  */
 
 static void
-VMMouseUnInit(InputDriverPtr drv, LocalDevicePtr local, int flags)
+VMMouseUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 {
    xf86Msg(X_INFO, "VMWARE(0): VMMouseUnInit\n");
 }
@@ -775,6 +806,10 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
    VMMousePrivPtr mPriv;
    unsigned char map[MSE_MAXBUTTONS + 1];
    int i;
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+   Atom btn_labels[MSE_MAXBUTTONS] = {0};
+   Atom axes_labels[2] = { 0, 0 };
+#endif
 
    pInfo = device->public.devicePrivate;
    pMse = pInfo->private;
@@ -790,12 +825,33 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
        */
       for (i = 0; i < MSE_MAXBUTTONS; i++)
 	 map[i + 1] = i + 1;
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+      btn_labels[0] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_LEFT);
+      btn_labels[1] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_MIDDLE);
+      btn_labels[2] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_RIGHT);
+      btn_labels[4] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_UP);
+      btn_labels[5] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_WHEEL_DOWN);
+      btn_labels[6] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_LEFT);
+      btn_labels[7] = XIGetKnownProperty(BTN_LABEL_PROP_BTN_HWHEEL_RIGHT);
+      /* other buttons are unknown */
+
+#ifdef ABS_VALUATOR_AXES
+      axes_labels[0] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_X);
+      axes_labels[1] = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_Y);
+#else
+      axes_labels[0] = XIGetKnownProperty(AXIS_LABEL_PROP_REL_X);
+      axes_labels[1] = XIGetKnownProperty(AXIS_LABEL_PROP_REL_Y);
+#endif /* ABS_VALUATOR_AXES */
+#endif
 
       InitPointerDeviceStruct((DevicePtr)device, map,
 			      min(pMse->buttons, MSE_MAXBUTTONS),
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+				btn_labels,
+#endif
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
 				miPointerGetMotionEvents,
-#else
+#elif GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 3
                                 GetMotionHistory,
 #endif
                                 pMse->Ctrl,
@@ -803,14 +859,57 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
 				miPointerGetMotionBufferSize()
 #else
                                 GetMotionHistorySize(), 2
+#endif
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+				, axes_labels
 #endif
                                 );
 
       /* X valuator */
-      xf86InitValuatorAxisStruct(device, 0, 0, -1, 1, 0, 1);
+#ifdef ABS_VALUATOR_AXES
+      xf86InitValuatorAxisStruct(device, 0,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+				axes_labels[0],
+#endif
+				0, 65535, 10000, 0, 10000
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Absolute
+#endif
+                                );
+#else
+      xf86InitValuatorAxisStruct(device, 0,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+				axes_labels[0],
+#endif
+				0, -1, 1, 0, 1
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Relative
+#endif
+                                );
+#endif
       xf86InitValuatorDefaults(device, 0);
       /* Y valuator */
-      xf86InitValuatorAxisStruct(device, 1, 0, -1, 1, 0, 1);
+#ifdef ABS_VALUATOR_AXES
+      xf86InitValuatorAxisStruct(device, 1,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+				axes_labels[1],
+#endif
+				0, 65535, 10000, 0, 10000
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Absolute
+#endif
+                                );
+#else
+      xf86InitValuatorAxisStruct(device, 1,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+				axes_labels[1],
+#endif
+				0, -1, 1, 0, 1
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+                                , Relative
+#endif
+                                );
+#endif
       xf86InitValuatorDefaults(device, 1);
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
       xf86MotionHistoryAllocate(pInfo);
@@ -846,8 +945,6 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
 		  return FALSE;
 	       } else {
 		  mPriv->vmmouseAvailable = TRUE;
-		  VMMouseClient_RequestAbsolute();
-		  mPriv->relative = FALSE;
 		  xf86Msg(X_INFO, "VMWARE(0): vmmouse enabled\n");
 	       }
 	    }
@@ -867,7 +964,8 @@ VMMouseDeviceControl(DeviceIntPtr device, int mode)
 	 VMMousePrivPtr mPriv = (VMMousePrivPtr)pMse->mousePriv;
 	 if( mPriv->vmmouseAvailable ) {
 	    VMMouseClient_Disable();
-	    mPriv->vmmouseAvailable = FALSE;
+            mPriv->vmmouseAvailable = FALSE;
+            mPriv->absoluteRequested = FALSE;
 	 }
 
 	 xf86RemoveEnabledDevice(pInfo);
@@ -917,6 +1015,17 @@ VMMouseReadInput(InputInfoPtr pInfo)
    pMse = pInfo->private;
    mPriv = pMse->mousePriv;
 
+   if (!mPriv->absoluteRequested) {
+      /*
+       * We can request for absolute mode, but it depends on
+       * host whether it will send us absolute or relative
+       * position.
+       */
+      VMMouseClient_RequestAbsolute();
+      mPriv->absoluteRequested = TRUE;
+      xf86Msg(X_INFO, "VMWARE(0): vmmouse enable absolute mode\n");
+   }
+
    /*
     * First read the bytes in input device to clear the regular PS/2 fd so
     * we don't get called again.
@@ -968,12 +1077,14 @@ VMMouseReadInput(InputInfoPtr pInfo)
 static void
 GetVMMouseMotionEvent(InputInfoPtr pInfo){
    MouseDevPtr pMse;
+   VMMousePrivPtr mPriv;
    int buttons, dx, dy, dz, dw;
    VMMOUSE_INPUT_DATA  vmmouseInput;
    int ps2Buttons = 0;
    int numPackets;
 
    pMse = pInfo->private;
+   mPriv = (VMMousePrivPtr)pMse->mousePriv;
    while((numPackets = VMMouseClient_GetInput(&vmmouseInput))){
       if (numPackets == VMMOUSE_ERROR) {
          VMMouseClient_Disable();
@@ -998,8 +1109,13 @@ GetVMMouseMotionEvent(InputInfoPtr pInfo){
       dy = vmmouseInput.Y;
       dz = (char)vmmouseInput.Z;
       dw = 0;
+      /*
+       * Get the per package relative or absolute information.
+       */
+      mPriv->isCurrRelative = vmmouseInput.Flags & VMMOUSE_MOVE_RELATIVE;
       /* post an event */
       pMse->PostEvent(pInfo, buttons, dx, dy, dz, dw);
+      mPriv->vmmousePrevInput = vmmouseInput;
    }
 }
 
@@ -1020,7 +1136,7 @@ GetVMMouseMotionEvent(InputInfoPtr pInfo){
  */
 
 static int
-VMMouseControlProc(LocalDevicePtr local, xDeviceCtl * control)
+VMMouseControlProc(InputInfoPtr pInfo, xDeviceCtl * control)
 {
    xf86Msg(X_INFO, "VMWARE(0): VMMouseControlProc\n");
    return (Success);
@@ -1042,11 +1158,13 @@ VMMouseControlProc(LocalDevicePtr local, xDeviceCtl * control)
  *----------------------------------------------------------------------
  */
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
 static void
-VMMouseCloseProc(LocalDevicePtr local)
+VMMouseCloseProc(InputInfoPtr pInfo)
 {
    xf86Msg(X_INFO, "VMWARE(0): VMMouseCloseProc\n");
 }
+#endif
 
 
 /*
@@ -1087,6 +1205,7 @@ VMMouseSwitchMode(ClientPtr client, DeviceIntPtr dev, int mode)
  *----------------------------------------------------------------------
  */
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
 static Bool
 VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v2,
 	     int v3, int v4, int v5, int *x, int *y)
@@ -1101,7 +1220,7 @@ VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v
    if (first != 0 || num != 2)
       return FALSE;
 
-   if(mPriv->relative) {
+   if(mPriv->isCurrRelative) {
       *x = v0;
       *y = v1;
    } else {
@@ -1117,17 +1236,10 @@ VMMouseConvertProc(InputInfoPtr pInfo, int first, int num, int v0, int v1, int v
    }
    return TRUE;
 }
+#endif
 
 
 #ifdef XFree86LOADER
-ModuleInfoRec VMMouseInfo = {
-    1,
-    "VMMOUSE",
-    NULL,
-    0,
-    VMMouseAvailableOptions,
-};
-
 
 /*
  *----------------------------------------------------------------------
@@ -1176,21 +1288,20 @@ VMMousePlug(pointer	module,
 	    int		*errmin)
 {
    static Bool Initialised = FALSE;
-   char *name;
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 7
    xf86LoaderReqSymLists(reqSymbols, NULL);
-
-   if (!Initialised) {
-      Initialised = TRUE;
-#ifndef REMOVE_LOADER_CHECK_MODULE_INFO
-      if (xf86LoaderCheckSymbol("xf86AddModuleInfo"))
 #endif
-	 xf86AddModuleInfo(&VMMouseInfo, module);
-   }
+
+   if (!Initialised)
+      Initialised = TRUE;
 
    xf86Msg(X_INFO, "VMWARE(0): VMMOUSE module was loaded\n");
    xf86AddInputDriver(&VMMOUSE, module, 0);
 
+#ifndef NO_MOUSE_MODULE
+{
+   char *name;
    /*
     * Load the normal mouse module as submodule
     * If we fail in PreInit later, this allows us to fall back to normal mouse module
@@ -1205,7 +1316,9 @@ VMMousePlug(pointer	module,
    if (!LoadSubModule(module, name, NULL, NULL, NULL, NULL, errmaj, errmin)) {
       LoaderErrorMsg(NULL, name, *errmaj, *errmin);
    }
-   xfree(name);
+   free(name);
+}
+#endif
 
    return module;
 }
@@ -1216,7 +1329,7 @@ static XF86ModuleVersionInfo VMMouseVersionRec = {
    MODINFOSTRING1,
    MODINFOSTRING2,
    XORG_VERSION_CURRENT,
-   VMMOUSE_MAJOR_VERSION, VMMOUSE_MINOR_VERSION, VMMOUSE_PATCHLEVEL,
+   PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR, PACKAGE_VERSION_PATCHLEVEL,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
@@ -1226,7 +1339,7 @@ static XF86ModuleVersionInfo VMMouseVersionRec = {
 /*
  * The variable contains the necessary information to load and initialize the module
  */
-XF86ModuleData vmmouseModuleData = {
+_X_EXPORT XF86ModuleData vmmouseModuleData = {
    &VMMouseVersionRec,
    VMMousePlug,
    VMMouseUnplug
